import sys
sys.path += ['.', './layers/', './utils/']

from hqm.layers.basiclayer import BasicLayer
from hqm.utils.sizes import size_conv_layer

import torch

class HybridLeNet5(torch.nn.Module):
    '''
        This class implements a quantum hybrid convolutional neural network based on LeNet-5.
        HybridLeNet5 is composed of classical convlutional block and hybrid quantum MLP.
        The size of the network output is defined by ou_dim.
    '''

    def __init__(self, qlayer : BasicLayer, in_shape : tuple, ou_dim : int) -> None:
        '''
            HybridLeNet5 constructor.  

            Parameters:  
            -----------  
            - qlayer : hqm.layers.basilayer.BasicLayer  
                hqm quantum layer to be stacked between two fully connected layers  
            - in_shape : tuple  
                tuple representing the shape of the input image (channels, widht, height)  
            - ou_dim : int  
                integer representing the output size of the hybrid model  
            
            Returns:  
            --------  
            Nothing, a HybridLeNet5 object will be created.    
        '''

        super().__init__()

        if len(in_shape) != 3: raise Exception(f"The parameter in_shape must be a tuple of four elements (channels, widht, height), found {in_shape}")
        if ou_dim < 1: raise Exception(f"The parameter ou_dim must be greater than 1, found {ou_dim}")
        
        c, w, h = in_shape
        
        c1 = 6
        self.conv_1    = torch.nn.Conv2d(in_channels=c, out_channels=c1, kernel_size=5, padding=2, stride=1)
        w1 = size_conv_layer(w, kernel_size=5, padding=2, stride=1)
        h1 = size_conv_layer(h, kernel_size=5, padding=2, stride=1)
        
        self.max_pool1 = torch.nn.MaxPool2d(kernel_size = (2,2), stride=(2,2))
        w2 = size_conv_layer(w1, kernel_size=2, padding=0, stride=2)
        h2 = size_conv_layer(h1, kernel_size=2, padding=0, stride=2)
        
        c2 = 16
        self.conv_2  = torch.nn.Conv2d(in_channels=c1, out_channels=c2, kernel_size=5,  stride=1)
        w3 = size_conv_layer(w2, kernel_size=5, padding=0, stride=1)
        h3 = size_conv_layer(h2, kernel_size=5, padding=0, stride=1)

        self.max_pool2 = torch.nn.MaxPool2d(kernel_size = (2,2), stride=(2,2))
        w4 = size_conv_layer(w3, kernel_size=2, padding=0, stride=2)
        h4 = size_conv_layer(h3, kernel_size=2, padding=0, stride=2)

        self.flatten_size = w4 * h4 * c2
        fc_2_size = int(self.flatten_size * 30 / 100)

        self.fc_1    = torch.nn.Linear(self.flatten_size, fc_2_size)
        self.fc_2    = torch.nn.Linear(fc_2_size, qlayer.n_qubits)
        self.qc_1    = qlayer.qlayer
        self.fc_3    = torch.nn.Linear(qlayer.n_qubits, ou_dim)
        self.relu    = torch.nn.ReLU()
        self.softmax = torch.nn.Softmax(dim=1)
    
    def forward(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method  

            Parameters:  
            -----------
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - out : torch.Tensor  
                output from the torch model  
        '''
        
        x = self.max_pool1(self.relu(self.conv_1(x)))
        x = self.max_pool2(self.relu(self.conv_2(x)))
        x = x.view(-1, self.flatten_size)
        x = self.relu(self.fc_1(x))
        x = self.relu(self.fc_2(x))
        x = self.relu(self.qc_1(x))
        x = self.fc_3(x)
        out = self.softmax(x)
        return out


from types import FunctionType
import pennylane as qml
import numpy as np
import sys

sys.path += ['.', './ciruits/']

from hqm.circuits.circuit import QuantumCircuit

class GaussianLikeNoiseGenerator(QuantumCircuit):
    '''
        This class implements a torch/keras quantum layer using a basic entangler
        circuit. 
    '''
    
    def __init__(self, location : float, scale : float, dev : qml.devices = None) -> None:
        '''
            GaussianLikeNoiseGenerator constructor.  

            Parameters:  
            -----------
            - location : float  
                location parameter translate the pdf, relative to the standard normal distribution
            - scale : float  
                scale parameter that stretches the pdf. The greater the magnitude, the greater the stretching. 
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set
                to 'default.qubit'. Recommendent qml.device("default.mixed", wires=1, shots=1000).
            
            Returns:  
            --------  
            Nothing, a GaussianLikeNoiseGenerator object will be created.  
        '''

        super().__init__(n_qubits=1, n_layers=1, dev=dev)

        self.shots     = dev.shots
        self.location  = location + 1/2
        self.scale     = scale * np.sqrt(self.shots/(1/2))
        self.circuit   = self.circ(self.dev)
    
    
    def generate_noise(self):
        '''
            GaussianLikeNoiseGenerator that generates on observable  

            Parameters:  
            -----------
            Nothing
            
            Returns:  
            --------  
            n : float
                gaussianlike quantum noise observable
        '''
        counts = self.circuit()
        n = self.scale*(counts[1]/self.shots-self.location)
        return n

    
    def generate_noise_array(self, shape:tuple):
        '''
            GaussianLikeNoiseGenerator that generates a matrix of observables

            Parameters:  
            -----------
            shape : tuple
                shape of the desired maxtrix of observables
            
            Returns:  
            --------  
            n : float
                gaussianlike quantum noise matrix of observables
        '''
        n = []
        for _ in range(np.prod(shape)): n.append(self.generate_noise())
        n = np.array(n)
    
        return n.reshape(shape)
    
    def circ(self, dev : qml.devices) -> FunctionType:
        '''
            GaussianLikeNoiseGenerator method that implements the quantum circuit.  

            Parameters:  
            -----------  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set  
                to 'default.qubit'   

            Returns:  
            --------  
            - qnode : qml.qnode  
                the actual PennyLane circuit   
        '''

        @qml.qnode(dev)
        def qnode() -> dict:
            '''
                PennyLane based quantum circuit composed of a Ry gate

                Parameters:  
                -----------  
                Nothing

                Returns:  
                --------  
                - counts : dict  
                    Sample from the supplied observable, with the number of shots determined from the dev.shots attribute of the corresponding device, 
                    returning the number of counts for each sample. If no observable is provided then basis state samples are returned directly 
                    from the device. Note that the output shape of this measurement process depends on the shots specified on the device. 
            '''

            qml.RY(np.pi/2, 0)
            return qml.counts(qml.PauliZ(0))
    
        return qnode


'''
    Work in progress
'''
from types import FunctionType
import pennylane as qml
import numpy as np
import sys

sys.path += ['.', './utils/']

from .circuit import QuantumCircuit

class BellmanCircuit(QuantumCircuit):
    '''
        This class implements a torch/keras quantum layer using the bellman
        circuit. 
    '''
    
    def __init__(self, n_qubits : int, n_layers : int, dev : qml.devices = None, encoding : str = 'angle') -> None:
        '''
            BellmanCircuit constructor.  

            Parameters:  
            -----------
            - n_qubits : int  
                number of qubits for the quantum circuit  
            - n_layers : int  
                number of layers for the quantum circuit  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set
                to 'default.qubit'  
            - encoding : str  
                string representing the type of input data encoding in quantum circuit, can be 'angle' or 'amplitude'
            
            Returns:  
            --------  
            Nothing, a BellmanCircuit object will be created.  
        '''
        
        super().__init__(n_qubits=n_qubits, n_layers=n_layers, dev=dev)

        if encoding not in ['angle', 'amplitude']: raise(f"encoding can be angle or amplitude, found {encoding}")
        
        self.encoding     = encoding
        self.weight_shape = {"weights": (n_layers, n_qubits, 3)}
        self.circuit      = self.circ(self.dev, self.n_qubits, self.n_layers, self.encoding)

    @staticmethod
    def circ(dev : qml.devices, n_qubits : int, n_layers : int, encoding : str) -> FunctionType:
        '''
            BellmanCircuit static method that implements the quantum circuit.  

            Parameters:  
            -----------  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set  
                to 'default.qubit'  
            - n_qubits : int  
                number of qubits for the quantum circuit  
            - n_layers : int  
                number of layers for the quantum circuit  
            - encoding : str  
                string representing the type of input data encoding in quantum circuit, can be 'angle' or 'amplitude'
            
            Returns:  
            --------  
            - qnode : qml.qnode  
                the actual PennyLane circuit   
        '''
        @qml.qnode(dev)
        def qnode(inputs : np.ndarray, weights : np.ndarray) -> list:
            '''
                PennyLane based quantum circuit composed of an angle embedding layer and a basic entangler
                layer.  

                Parameters:  
                -----------  
                - inputs : np.ndarray  
                    array containing input values (can came from previous torch/keras layers or quantum layers)  
                - weights : np.ndarray  
                    array containing the weights of the circuit that are tuned during training, the shape of this
                    array depends on circuit's layers and qubits.   
                
                Returns:  
                --------  
                - measurements : list  
                    list of values measured from the quantum circuits  
            '''

            if encoding == 'angle':     qml.AngleEmbedding(inputs, wires=range(n_qubits))
            if encoding == 'amplitude': qml.AmplitudeEmbedding(features=inputs, wires=range(n_qubits), normalize=True)

            for n in range(n_layers):
                qml.Hadamard(wires=0)
                
                # CNOT gates
                for q in range(1, n_qubits):
                    qml.CNOT(wires=[q-1, q])

                # Rot gates in the middle
                for q in range(n_qubits):
                    qml.Rot(weights[n, q, 0], weights[n, q, 1], weights[n, q, 2], wires=q)

                # CNOT gates
                for q in range(1, n_qubits):
                    qml.CNOT(wires=[n_qubits - q-1, n_qubits - q])

            measurements = [qml.expval(qml.PauliZ(wires=i)) for i in range(n_qubits)]
            return measurements
    
        return qnode

class RealAmplitudesCircuit(QuantumCircuit):
    '''
        This class implements a torch/keras quantum layer using the real amplitude circuit. 
    '''
    
    def __init__(self, n_qubits : int, n_layers : int, dev : qml.devices = None, encoding : str = 'angle') -> None:
        '''
            RealAmplitudesCircuit constructor.  

            Parameters:  
            -----------
            - n_qubits : int  
                number of qubits for the quantum circuit  
            - n_layers : int  
                number of layers for the quantum circuit  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set
                to 'default.qubit'  
            - encoding : str  
                string representing the type of input data encoding in quantum circuit, can be 'angle' or 'amplitude'
            
            Returns:  
            --------  
            Nothing, a RealAmplitudesCircuit object will be created.  
        '''
        super().__init__(n_qubits=n_qubits, n_layers=n_layers, dev=dev)

        if encoding not in ['angle', 'amplitude']: raise(f"encoding can be angle or amplitude, found {encoding}")
        
        self.encoding     = encoding
        self.weight_shape = {"weights": (n_layers, n_qubits, 3)}
        self.circuit      = self.circ(self.dev, self.n_qubits, self.n_layers, self.encoding)

    @staticmethod
    def circ(dev : qml.devices, n_qubits : int, n_layers : int, encoding : str) -> FunctionType:
        '''
            RealAmplitudesCircuit static method that implements the quantum circuit.  

            Parameters:  
            -----------  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set  
                to 'default.qubit'  
            - n_qubits : int  
                number of qubits for the quantum circuit  
            - n_layers : int  
                number of layers for the quantum circuit  
            - encoding : str  
                string representing the type of input data encoding in quantum circuit, can be 'angle' or 'amplitude'
            
            Returns:  
            --------  
            - qnode : qml.qnode  
                the actual PennyLane circuit   
        '''
        @qml.qnode(dev)
        def qnode(inputs : np.ndarray, weights : np.ndarray) -> list:
            '''
                PennyLane based quantum circuit composed of an angle embedding layer and a basic entangler
                layer.  

                Parameters:  
                -----------  
                - inputs : np.ndarray  
                    array containing input values (can came from previous torch/keras layers or quantum layers)  
                - weights : np.ndarray  
                    array containing the weights of the circuit that are tuned during training, the shape of this
                    array depends on circuit's layers and qubits.   
                
                Returns:  
                --------  
                - measurements : list  
                    list of values measured from the quantum circuits  
            '''

            if encoding == 'angle':     qml.AngleEmbedding(inputs, wires=range(n_qubits))
            if encoding == 'amplitude': qml.AmplitudeEmbedding(features=inputs, wires=range(n_qubits), normalize=True)

            for q in range(n_qubits): qml.Hadamard(wires=q)

            for n in range(n_layers):
                for q1 in range(0, n_qubits):
                    for q2 in range(0, n_qubits):
                        if q1 < q2:
                            qml.CNOT(wires=[q2, q1])
                
                # Rot gates in the middle
                for q in range(n_qubits):
                    qml.Rot(weights[n, q, 0], weights[n, q, 1], weights[n, q, 2], wires=q)


            measurements = [qml.expval(qml.PauliZ(wires=i)) for i in range(n_qubits)]
            return measurements
    
        return qnode


import sys

sys.path += ['.', './utils/', '/circuits/']

from hqm.circuits.circuit import QuantumCircuit
from hqm.utils.aiinterface import AIInterface


class BasicLayer:
    '''
        Basic Quantum Layer
    '''

    def __init__(self, qcircuit : QuantumCircuit, aiframework : str) -> None:
        '''
        BasicLayer constructor.  

        Parameters:  
        -----------  
        - qcircuit : hqm.circuits.circuit.QuantumCircuit  
            QuantumCircuit object to be embedded into the quantum layer
        - aiframework : str    
            string representing the AI framework in use, can be 'torch' or 'keras'. This will create  
            a compatible trainable layer for the framework.   

        Returns:    
        --------     
        Nothing, a BasicLayer object will be created.  
        '''
        
        if aiframework not in ['torch', 'keras']: raise Exception(f"Accepted values for framerwork are 'torch' or 'keras', found {aiframework}")

        self.aiframework = aiframework
        self.n_qubits    = qcircuit.n_qubits
        self.qlayer      = AIInterface.network_layer(
                                circuit      = qcircuit.circuit, 
                                weight_shape = qcircuit.weight_shape, 
                                n_qubits     = qcircuit.n_qubits, 
                                aiframework  = self.aiframework
                            )


import sys
sys.path += ['.', './hqm/']

from hqm.circuits.customcircuits import BellmanCircuit, RealAmplitudesCircuit

import numpy as np

def test_BasicEntangledCircuit_angle_output():
    n_qubits = 2
    n_layers = 1
    circ     = BellmanCircuit(n_qubits=n_qubits, n_layers=n_layers, encoding='angle').circuit
    cinput   = np.random.randn(n_qubits)
    cweights = np.random.randn(n_layers, n_qubits, 3)
    out      = circ(cinput, cweights)
    assert len(out) == n_qubits


def test_BasicEntangledCircuit_amplitude_output():
    n_qubits = 2
    n_layers = 1
    circ     = BellmanCircuit(n_qubits=n_qubits, n_layers=n_layers, encoding='amplitude').circuit
    cinput   = np.random.randn(2**n_qubits)
    cweights = np.random.randn(n_layers, n_qubits, 3)
    out      = circ(cinput, cweights)
    assert len(out) == n_qubits

def test_RealAmplitudesCircuitt_angle_output():
    n_qubits = 2
    n_layers = 1
    circ     = RealAmplitudesCircuit(n_qubits=n_qubits, n_layers=n_layers, encoding='angle').circuit
    cinput   = np.random.randn(n_qubits)
    cweights = np.random.randn(n_layers, n_qubits, 3)
    out      = circ(cinput, cweights)
    assert len(out) == n_qubits

def test_RealAmplitudesCircuit_amplitude_output():
    n_qubits = 2
    n_layers = 1
    circ     = RealAmplitudesCircuit(n_qubits=n_qubits, n_layers=n_layers, encoding='amplitude').circuit
    cinput   = np.random.randn(2**n_qubits)
    cweights = np.random.randn(n_layers, n_qubits, 3)
    out      = circ(cinput, cweights)
    assert len(out) == n_qubits


import sys
sys.path += ['.', './hqm/']

from hqm.noise.gaussianlike import GaussianLikeNoiseGenerator
import pennylane as qml

def test_GaussianLikeNoiseGenerator_output():
    gen = GaussianLikeNoiseGenerator(location=0, scale=1, dev=qml.device("default.mixed", wires=1, shots=1000))
    n = gen.generate_noise_array((64,30))
    assert n.shape == (64,30)



"""
   .. include:: ../README.md
"""

from .classification import *
from .regression import *
from .encoding import *
from .circuits import *
from .layers import *
from .noise import *
from .utils import *


from typing import TypeVar, Generic
import pennylane as qml
import warnings

quantumcircuit = TypeVar("quantumcircuit")
class QuantumCircuit(Generic[quantumcircuit]):
    '''
        Basic QuantumCircuit object.
    '''
    def __init__(self, n_qubits : int, n_layers : int, dev : qml.devices = None) -> None:
        '''
            QuantumCircuit parent class.  

            Parameters:  
            -----------
            - n_qubits : int  
                number of qubits for the quantum circuit  
            - n_layers : int  
                number of layers for the quantum circuit  
            - dev : qml.device  
                PennyLane device on wich run quantum operations (dafault : None). When None it will be set
                to 'default.qubit'  
            
            Returns:  
            --------  
            Nothing, a QuantumCircuit object will be created.  
        '''
        
        # Checking for exceptions  
        if n_qubits < 1: raise Exception(f"Number of qubits must be greater or equal than 1, found {n_qubits}")
        if n_layers < 1: raise Exception(f"Number of layers must be greater or equal than 1, found {n_layers}")

        # Set dev to 'default.qubit' if dev is None  
        if dev is None: 
            dev = qml.device("default.qubit", wires=n_qubits)
            warnings.warn(f"Dev has been set to None, setting it to {dev}")

        self.n_qubits     = n_qubits
        self.n_layers     = n_layers
        self.dev          = dev